"""
Information about the package.
"""

from datetime import date

__version__ = "0.1.2"
__author__ = u"Benjamin Vial"
__author_email__ = "benjamin.vial84@gmail.com"
__copyright__ = u"Copyright (c) 2020-{}, {} <{}>".format(
    date.today().year, __author__, __author_email__
)
__website__ = "https://gyptis.gitlab.io"
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Development Status :: 1 - Planning"
__audience__ = "Intended Audience :: Science/Research"
__operating_system__ = "Operating System :: OS Independent"
__programming_language__ = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
]
__topic__ = [
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Physics",
]
__description__ = "Computational Photonics in Python"
